function result=Cra(x,y,param)
% result=Cra(x,y,param)
%
% parameter:
%
% x                    (n,2) matrix of fixations or column vector of regions
% y                    (n,2) matrix of fixations or column vector of regions
% param.linelength     threshold linelength
% param.radius         e.g. 64, for categorial use 0.1
%
% result:
%
% result.n             n (length of x,y)
% result.recmat        recurrence matrix
% result.nrec          number of recurrences
% result.rec           percentage of recurrences
% result.rr            recurrence rate
% result.det           determinism
% result.meanline      meanline
% result.maxline       maxline
% result.ent           entropy
% result.relent        relative entropy
% result.lam           laminarity
% result.tt            TT
% result.corm          center of recurrence mass
%
% Author               Walter F. Bischof
% Date                 September 2011 - June 2012
%                      Simplified version: August 2015

    % Default return values
    
    result.n=NaN;
    result.recmat=[];
    result.nrec=NaN;
    result.rec=NaN;
    result.rr=[];
    result.det=NaN;
    result.meanline=NaN;
    result.maxline=NaN;
    result.ent=NaN;
    result.relent=NaN;
    result.lam=NaN;
    result.tt=NaN;
    result.corm=NaN;

    n=size(x,1);

    % Are they the same length
    if size(x,1) ~= size(y,1)
        return
    end
    
    result.n=n;
    
    if n<=1
        return
    end
    
    % Compute distance matrix. This uses most of the computation time.

    distanceMatrix=ComputeDistanceMatrix(x,y);

    % Recurrence matrix

    recurrenceMatrix=distanceMatrix<=param.radius;
    result.recmat=recurrenceMatrix;
    
    % Diagonals of recurrence matrix, excluding main diagonal

    [recurrenceDiagonals,diagonals]=spdiags(recurrenceMatrix);

    % Recurrence measures

    % Global measures: %REC
    nRec=sum(sum(recurrenceMatrix));
    result.nrec=nRec;
    result.rec=100.0*nRec/(n*n);

    % Recurrence rate
    result.rr=zeros(2,2*n-1);
    result.rr(1,:)=-n+1:n-1;
    result.rr(2,diagonals+n)=sum(recurrenceDiagonals,1);

    % Diagonal line measures: %DET,meanLine,maxLine,TND,ENT
    
    nDiagonals=length(diagonals);
    thresholdedDiagonals=[];
    for i=1:nDiagonals
        d=[0; recurrenceDiagonals(:,i); 0];
        dchange=diff(d);
        dlength=find(dchange==-1)-find(dchange==1);
        dlength=dlength(dlength>=param.linelength);
        thresholdedDiagonals=[thresholdedDiagonals; dlength];
    end

    if ~isempty(thresholdedDiagonals)
        result.det=100*sum(thresholdedDiagonals)/nRec;
        result.meanline=mean(thresholdedDiagonals);
        result.maxline=max(thresholdedDiagonals);
        [result.ent,result.relent]=Entropy(thresholdedDiagonals,param.linelength,result.maxline);
        result.corm=100*sum(sum(recurrenceDiagonals)'.*diagonals)/((n-1)*nRec);
    end
    
    % Horizontal+vertical line measures: %LAM, TT
    % #(verticals + horizontals) in upper triangle = 
    % #(verticals) in upper + #(verticals) in lower triangle
    % Set diagonal to zero to avoid counting over the diagonal
    
    thresholdedVerticals=[];
    paddedRM=[zeros(n,1) recurrenceMatrix-eye(n) zeros(n,1)];
    % In CRA, diagonal elements can be zero 
    paddedRM=max(paddedRM,0);
    for i=1:n
        v=paddedRM(i,:);
        vchange=diff(v);
        vlength=find(vchange==-1)-find(vchange==1);
        vlength=vlength(vlength>=param.linelength);
        thresholdedVerticals=[thresholdedVerticals vlength];
    end
    
    if ~isempty(thresholdedVerticals)
        result.lam=100*sum(thresholdedVerticals)/sum(sum(recurrenceMatrix));
        result.tt=mean(thresholdedVerticals);
    end
end

function [entropy,relativeEntropy]=Entropy(a,minLine,maxLine)
    [p,~]=hist(a,unique(a));
    p=p/sum(p);
    entropy=-sum(p.*log2(p));
    if maxLine==minLine
        relativeEntropy=NaN;
    else
        relativeEntropy=entropy/log2(maxLine-minLine+1);
    end
end

function distanceMatrix=ComputeDistanceMatrix(x,y)
% Compute the distance matrix between xSeq and ySeq.
distanceMatrix=pdist2(x,y); 
end
