function result=Rqa(x,param)
% result=Rqa(x,param)
%
% parameter:
%
% x                    (n,2) matrix of x,y fixations or
%                      (n,1) column vector of region numbers
% param.linelength     threshold linelength, e.g. 2
% param.radius         e.g. 64. For categorical data use 0.1
%
% result:
%
% result.n             n (length of x)
% result.recmat        recurrence matrix
% result.nrec          number of recurrences
% result.rec           percentage of recurrences
% result.det           determinism
% result.meanline      meanline
% result.maxline       maxline
% result.ent           entropy
% result.relent        relative entropy
% result.trend         trend
% result.lam           laminarity
% result.tt            TT
% result.corm          center of recurrence mass
%
% Author               Walter F. Bischof
% Date                 September 2011 - September 2012
%                      Simplified version: August 2015

    % Default return values
    
    result.n=NaN;
    result.recmat=NaN;
    result.nrec=NaN;
    result.rec=NaN;
    result.det=NaN;
    result.meanline=NaN;
    result.maxline=NaN;
    result.ent=NaN;
    result.relent=NaN;
    result.lam=NaN;
    result.tt=NaN;
    result.trend=NaN;
    result.corm=NaN;
        
    n=size(x,1);
    result.n=n;
    
    % Compute distance matrix.

    distanceMatrix=ComputeDistanceMatrix(x,x);
    nTriangle=n*(n-1)/2;

    % Recurrence matrix

    recurrenceMatrix=distanceMatrix<=param.radius;
    result.recmat=recurrenceMatrix;
    
    % Diagonal not needed for most measures
    partialRecurrenceMatrix=triu(recurrenceMatrix,1);

    % Diagonals of recurrence matrix, excluding main diagonal

    [recurrenceDiagonals,diagonals]=spdiags(partialRecurrenceMatrix);

    % Recurrence measures

    % Global measures: %REC
    nRec=sum(sum(partialRecurrenceMatrix));
    result.nrec=nRec;
    result.rec=100.0*nRec/nTriangle;
    
    if nRec==0
        return
    end

    % Diagonal line measures: %DET,meanLine,maxLine,TND,ENT
    
    nDiagonals=length(diagonals);
    thresholdedDiagonals=[];
    for i=1:nDiagonals
        d=[0; recurrenceDiagonals(:,i); 0];
        dchange=diff(d);
        dlength=find(dchange==-1)-find(dchange==1);
        dlength=dlength(dlength>=param.linelength);
        thresholdedDiagonals=[thresholdedDiagonals; dlength];
    end

    if ~isempty(thresholdedDiagonals)
        result.det=100*sum(thresholdedDiagonals)/nRec;
        result.meanline=mean(thresholdedDiagonals);
        result.maxline=max(thresholdedDiagonals);
        [result.ent,result.relent]=Entropy(thresholdedDiagonals,param.linelength,result.maxline);

        sumRD=100*sum(recurrenceDiagonals)'./(n-diagonals);
        predictor=1:n; predictor=predictor';
        response=zeros(n,1);
        response(diagonals)=sumRD;
        % eliminate outermost 10%, acccording to Chuck Webber
        nSubset=floor(0.9*n);
        predictor=predictor(1:nSubset);
        response=response(1:nSubset);
        result.trend=polyfit(predictor,response,1);
        result.trend=1000*result.trend(1);  
        result.corm=100*sum(sum(recurrenceDiagonals)'.*diagonals)/((n-1)*nRec);
    end
    
    % horizontal+vertical line measures: %LAM, TT.
    % These are slightly different from Zbilut & Webber and Marwan
    % #(verticals + horizontals) in upper triangle = 
    % #(verticals) in upper + #(verticals) in lower triangle
    % Set diagonal to zero to avoid counting over the diagonal
    
    thresholdedVerticals=[];
    paddedRM=[zeros(n,1) recurrenceMatrix-eye(n) zeros(n,1)];
    % In CRA, diagonal elements can be zero 
    paddedRM=max(paddedRM,0);
    for i=1:n
        v=paddedRM(i,:);
        vchange=diff(v);
        vlength=find(vchange==-1)-find(vchange==1);
        vlength=vlength(vlength>=param.linelength);
        thresholdedVerticals=[thresholdedVerticals vlength];
    end
    
    if ~isempty(thresholdedVerticals)
        result.lam=100*sum(thresholdedVerticals)/sum(sum(recurrenceMatrix));
        result.tt=mean(thresholdedVerticals);
    end            
end

function [entropy,relativeEntropy]=Entropy(a,minLine,maxLine)
    [p,~]=hist(a,unique(a));
    p=p/sum(p);
    entropy=-sum(p.*log2(p));
    if maxLine==minLine
        relativeEntropy=NaN;
    else
        relativeEntropy=entropy/log2(maxLine-minLine+1);
    end
end

function distanceMatrix=ComputeDistanceMatrix(x,y)
% Compute the distance matrix between x and y.
distanceMatrix=pdist2(x,y); 
end
